#------------------------------------------------------------------
# $Copyright: Copyright (c) 2022 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
# $Header: /project/isv-agent/VProAgents/SAPHDB/SAPHDBSR.py,v 1.0 2022/02/11 11:30:42 sbirhade Exp $ 
#
# $Author: sbirhade $
# $Date: 2022/02/11 11:30:42 $
# $Revision: 1.0 $
#
# Provides:                            SAPHDB
# Author:                              sbirhade
#------------------------------------------------------------------


from hdb_ha_dr.client import HADRBase
import os
import socket
import tempfile
import time

class SAPHDBSR (HADRBase):

    def __init__(self, *args, **kwargs):
        # delegate construction to base class
        super(SAPHDBSR, self).__init__(*args, **kwargs)

        self._physical_hostname = socket.gethostname()
        self._takeover_group = self.config.get("takeover_group")
        

    def postTakeover(self, rc, **kwargs):
        """
        Hook description:
        
        * time of call: as soon as all services with a volume return from their 
                        assign-call (open SQL port)
        * caller: the master host
        * landscape: called only once on the master 
        * behavior upon failure: error trace is written

        @param rc: the return code of the actual takeover process; 0=success, 
                   1=waiting for forced takeover, 2=failure
        @type rc: int
        @param **kwargs: place holder for later usage (new parameters) to 
                         keep the interface stable
        @type **kwargs: dict  
        @return: information about success
        @rtype: int
        """

        self.tracer.info("%s.postTakeover method called with rc=%s" % (self.__class__.__name__, rc))
        
        if rc == 0:
            return_value = 0 if self.update_status(self._physical_hostname) else 1
        else:
            self.tracer.warning("Take-over did not complete successfully with rc: [%s]" % rc)
            return_value = 0 if self.update_status(None) else 1

        return return_value


    def update_status(self, hostname, sync_status = 0):
        """
        Function Description:

        * This function is invoked to udpate primary hostname & syn staus.

        @param hostname: Host name to be updated in the GlobalUserStr attribue of takeover SG
        @type hostname: str
        @param sync_status: Sync status of the HANA DataBase
                            0: not in-sync
                            1: in-sync
        @type sync_status: int
        @return: Information about success/failure
                 True: Sucess
                 False: Failure
        @type: Bool
        """

        cluster_home = "/opt/VRTS/bin"
        cluster_hostname = None
        cluster_sync_status = None
        cluster_hostname = None
        return_value = 0

        cmd = "{}/hagrp -value {} UserStrGlobal".format(cluster_home, self._takeover_group)
        try:
             rc, result = self.run_command(cmd)
        except Exception as err:
            self.tracer.error("Command [%s] failed with exception [%s]" % (cmd, err))
            return return_value
        else:
            if rc == 0:
                if result[0] != "\n":
                    values = result[0].split(";")
                    cluster_hostname = values[0]
                    cluster_sync_status = values[1]
                self.tracer.info("Command [%s] executed successfully with output [%s]" % (cmd, result))
            else:
                self.tracer.error("Failed reading cluster configuration with error code: [%s] and output: [%s]" % (rc, result))
                return False

        if hostname == None:
            if cluster_hostname != None:
                hostname = cluster_hostname
            else:
                self.trace.warning("Attribute UserStrGlobal is not set and takeover has failed")
                self.trace.warning("Not setting UserStrGlobal for Group [%s]" %(self._takeover_group))
                return False

        return_value = False
        self.tracer.debug("Opening cluster for writing")
        count = 10
        while count > 0:
            cmd = "haconf -makerw"
            cmd = "{}/{}".format(cluster_home , cmd)
            try:
                rc, result = self.run_command(cmd)
            except Exception as err:
                self.tracer.error("Command [%s] failed with exception [%s]" % (cmd, err))
                return False
		
            if rc == 0 or result[0] == "VCS WARNING V-16-1-10364 Cluster already writable.\n":
                if hostname != cluster_hostname or sync_status != int(cluster_sync_status):
                    cmd = "hagrp -modify {} UserStrGlobal \"{};{};{}\"".format(
                                         self._takeover_group,
                                         hostname,
                                         sync_status,
                                         time.strftime("%Y-%m-%d_%H:%M:%S", time.localtime())
                                         )
                    cmd = "{}/{}".format(cluster_home, cmd)
                    try:
                       rc, result = self.run_command(cmd)
                    except Exception as err:
                           self.tracer.error("Command [%s] failed with exception %s" % (cmd, err))
                           break
                    else:
                       if rc != 0:
                           self.tracer.warning("Command [%s] failed with error [%s] and exit code [%s]" % (cmd, rc, result))
                           self.tracer.info("Sleeping for 1 sec & then retrying...")
                           time.sleep (1)
                           count = count - 1
                           continue
                       else:
                           self.tracer.info("Command cmd [%s] is executed successfully" % cmd)
                           return_value = True
                           break
                else:
                    self.tracer.info("Not setting UserStrGlobal")
                    return_value = True
                    break

            else:
                self.tracer.warning("Command [%s] failed with error [%s] and exit code [%s]" % (cmd, rc, result))
                self.tracer.info("Sleeping for 1 sec & then retrying...")
                time.sleep (1)
                count = count - 1
                continue

        if return_value == False: 
            self.tracer.error("Unable to update UserStrGlobal attribute of takeover SG [%s]. Exiting..." % self._takeover_group)
            return return_value
        
        return_value = False
        count = 10
        cmd = "haconf -dump -makero"
        cmd = "{}/{}".format(cluster_home, cmd)
        self.tracer.debug("Closing cluster configuration")
        while count > 0:
            try:
                rc, result = self.run_command(cmd)
            except Exception as err:
                self.tracer.error("Command [%s] failed with exception [%s]" % (cmd, err))
                break
            else:
                if rc == 0 or result[0] == "VCS WARNING V-16-1-10369 Cluster not writable.\n":
                    self.tracer.info("Command [%s] is executed successfully" % cmd)
                    return_value = True
                    break
                else:
                    self.tracer.error("Command [%s] failed with error code [%s] and message [%s]. Retrying..." % (cmd, rc, result))
                    time.sleep (1)
                    count = count - 1
     
        return return_value


    def run_command(self, cmd):
        """
        Function Description:

        * This function is invoked to run a command & return error code with output/error.

        @param cmd: Command to be executed.
        @type cmd: str
        @return: Exit code and output/error
        @type: Tuple
        """

        ftmp = tempfile.NamedTemporaryFile(suffix='.out', prefix='SAPHANASR', delete=False)
        fpath = ftmp.name
        ftmp.close()
        result = []

        self.tracer.debug("Running command [%s]" % cmd)
        cmd = "{} > {} 2>&1".format(cmd, fpath)
        try:
            rc = os.system(cmd)
            rc  = 0 if rc is None else os.WEXITSTATUS(rc)
        except Exception as err:
            os.remove(fpath)
            self.tracer.error("Command [%s] failed with exception [%s]" % (cmd, err))
            return rc, result
        
        try:
            with open(fpath, 'r') as file:
                result = file.readlines()
        except Exception as err:
            os.remove(fpath)
            self.tracer.error("Open failed with exception [%s]" % err)
            return 1, result
        
        os.remove(fpath)
        return rc, result


    def srConnectionChanged(self, ParamDict, **kwargs):
        """
        Hook description:

        * time of call: as soon as one of the replicating services loses or 
                        (re-)establishes the system replication connection
        * caller: master node on primary site
        * landscape: called only once on the master node on primary site
        * behavior upon failure: error trace is written
        * Possible return codes:
        * 0: Ok - continue processing
        * 1: Block - Further SAP HANA processing is blocked. Every 5 sec. there 
                     will be a retry to call this hook
        * If an HA/DR Provider shall not block SAP HANA processing or if a blocking
                     situation caused by an HA/DR Provider shall (temporarily) be
                     resolved 
        * use 'hdbnsutil -sr_blockonconnectionchanged --disable'
        @param parameters: dict of parameters {hostname:string, port:string,
                           database:string, status:int, database_status:int, 
                           system_status:int, timestamp:string, is_in_sync:bool, reason:string}
        @type parameters: dict
        @param **kwargs: place holder for later usage (new parameters) to keep the
                         interface stable
        @type **kwargs: dict
        @return: information about success (0 = continue transaction, 1 = halt
                 further transactions)
        @rtype: int

        * parameters:
        * -- hostname: host where the service is running
        * -- port: service's port
        * -- database: service's tenant database (MDC)
        * -- status: service replication status (10: NoHSR, 11: Error, 12: Unknown, 13: Initializing, 14: Syncing, 15: Active)
        * -- database_status: tenant database replicating status (10: NoHSR, 11: Error, 12: Unknown, 13: Initializing, 14: Syncing, 15: Active)
        * -- system_status: HANA database overall replicating status (10: NoHSR, 11: Error, 12: Unknown, 13: Initializing, 14: Syncing, 15: Active)
        * -- timestamp: date and time of the event
        * -- is_in_sync: true if service is in sync
        * -- reason: additional details (e.g. 'Starting', 'Stopping')
        * -- siteName: name of the replicating secondary site

        """
        self.tracer.info("srConnectionChanged method called with Dict=%s" % (ParamDict))
        system_status = ParamDict.get("system_status")
        reason = ParamDict.get("reason")

        return_value = 0
        if system_status == 15:
            return_value = 0 if self.update_status(self._physical_hostname, 1) else 1
        else:
            if reason != "Stopping":
                return_value = 0 if self.update_status(None, 0) else 1            

        return return_value



    def preTakeover(self, isForce, **kwargs):
        """
        Hook description:
        
        * time of call: as soon as the hdbnsutil -sr_takeover command is issued
        * caller: the master host
        * landscape: called only once on the master 
        * behavior upon failure: nameserver aborts, takeover is cancelled

        @param isForce: flag if it is a normal or forced takeover (as of today, 
                        takeover is always forced regardless of the value of the 
                        flag)
        @type isForce: bool
        @param **kwargs: place holder for later usage (new parameters) to 
                         keep the interface stable
        @type **kwargs: dict  
        @return: information about success
        @rtype: int
        """

        self.tracer.info("preTakeover method called with isForce=%s" % (isForce))

        return_value = 0 if self.update_status(self._physical_hostname, 0) else 1            
        return return_value
